"""Command Line Interface for Airshare."""


import click
import os
import pyperclip


from .utils import is_file_copyable, get_clipboard_paths
from .sender import send, send_server
from .receiver import receive, receive_server


@click.command(name="airshare")
@click.argument("code", nargs=1)
@click.option("-p", "--port", type=int, default=80, help="""
Specify the port number to host a sending or receiving server (defaults to 80).
""")
@click.option("-t", "--text", type=str, help="""
Send (serve) text content. For multiple words, enclose within quotes.
""")
@click.option("-u", "--upload", is_flag=True, help="""
Host a receiving server or upload file(s) to one.
""")
@click.option("-cs", "--clip-send", is_flag=True, help="""
Send (serve) clipboard contents as text.
""")
@click.option("-cr", "--clip-receive", is_flag=True, help="""
Receive served content and also copy into clipboard (if possible).
""")
@click.option("-fp", "--file-path", is_flag=True, help="""
Send files whose paths have been copied to the clipoard.
""")
@click.argument("files", nargs=-1)
@click.help_option()
@click.version_option(version=None, prog_name="Airshare")
def main(code, port, text, upload, clip_send, clip_receive, file_path, files):
    r"""Airshare - an easy way to share content in a local network.

    CODE - An identifying code for Airshare.

    FILES - File(s) or directories to send.
    """
    files = get_clipboard_paths() if file_path else files
    if text:
        try:
            send_server(code=code, text=text, port=port)
        except KeyboardInterrupt:
            exit(0)
    if clip_send:
        try:
            send_server(code=code, text=pyperclip.paste(), port=port)
        except KeyboardInterrupt:
            exit(0)
    if clip_receive:
        content = receive(code=code)
        if os.path.exists(content):
            if is_file_copyable(content):
                with open(content, "r") as f:
                    pyperclip.copy(f.read())
                print("File copied to clipboard!")
            else:
                print("This file cannot be copied to the clipboard!")
        else:
            pyperclip.copy(content)
        return
    if len(files):
        if upload:
            send(code=code, file=list(files))
            return
        else:
            try:
                send_server(code=code, file=files, port=port)
            except KeyboardInterrupt:
                exit(0)
    else:
        if upload:
            try:
                receive_server(code=code, port=port)
            except KeyboardInterrupt:
                exit(0)
        else:
            receive(code=code)


if __name__ == "__main__":
    main(prog_name="Airshare")



"""Exceptions defined for Airshare."""


__all__ = ["CodeExistsError", "CodeNotFoundError", "IsNotReceiverError",
           "IsNotSenderError"]


class AirshareError(Exception):
    """Base class for all Airshare-related exceptions."""


class CodeExistsError(AirshareError):
    r"""Exception to be raised when Airshare code already exists.

    To be raised when trying to create or serve an Airshare server, but
    one already exists with that identifying code.
    """
    def __init__(self, code):
        self.code = code

    def __str__(self):
        return "CodeExistsError: The Airshare `" + self.code \
               + ".local` already exists!"


class CodeNotFoundError(AirshareError):
    r"""Exception to be raised when the Airshare code is not found.

    To be raised when trying to download from or upload to an Airshare,
    but none exists with that identifying code.
    """
    def __init__(self, code):
        self.code = code

    def __str__(self):
        return "CodeNotFoundError: The Airshare `" + self.code \
               + ".local` was not found!"


class IsNotReceiverError(AirshareError):
    r"""Exception to be raised when trying to upload to a non-receiver.

    To be raised when trying to upload to an Airshare when it exists but
    is not a receiver.
    """
    def __init__(self, code):
        self.code = code

    def __str__(self):
        return "IsNotReceiverError: The Airshare `" + self.code \
               + ".local` is not an Upload Receiver!"


class IsNotSenderError(AirshareError):
    r"""Exception to be raised when trying to receive from a non-sender.

    To be raised when trying to download from an Airshare when it exists but
    is not a sender.
    """
    def __init__(self, code):
        self.code = code

    def __str__(self):
        return "IsNotSenderError: The Airshare `" + self.code \
               + ".local` is not a Text or File Sender!"



r"""QR Code Generator for Python
Copyright (c) 2012 Kazuhiko Arase
URL: http://www.d-project.com/
Licensed under the MIT license:
http://www.opensource.org/licenses/mit-license.php
The word 'QR Code' is registered trademark of
DENSO WAVE INCORPORATED
http://www.denso-wave.com/qrcode/faqpatent-e.html
Modified by Airshare Developers:
  - added `printQr()` to `QRCode`
"""


import colorama
from termcolor import colored, cprint


colorama.init()


class QRCode:

    PAD0 = 0xEC
    PAD1 = 0x11

    def __init__(self):
        self.typeNumber = 1
        self.errorCorrectLevel = ErrorCorrectLevel.H
        self.qrDataList = []
        self.modules = []
        self.moduleCount = 0

    def getTypeNumber(self):
        return self.typeNumber

    def setTypeNumber(self, typeNumber):
        self.typeNumber = typeNumber

    def getErrorCorrectLevel(self):
        return self.errorCorrectLevel

    def setErrorCorrectLevel(self, errorCorrectLevel):
        self.errorCorrectLevel = errorCorrectLevel

    def clearData(self):
        self.qrDataList = []

    def addData(self, data):
        self.qrDataList.append(QR8BitByte(data))

    def getDataCount(self):
        return len(self.qrDataList)

    def getData(self, index):
        return self.qrDataList[index]

    def isDark(self, row, col):
        return (self.modules[row][col] if self.modules[row][col] is not None
                else False)

    def getModuleCount(self):
        return self.moduleCount

    def printQr(self):
        WHITE = colored("  ", "white", "on_white")
        BLACK = "  "
        size = self.getModuleCount() + 2
        print()
        for row in range(size):
            for col in range(size):
                if row == 0 or col == 0 or col == size - 1 or row == size - 1:
                    cprint(WHITE, end="")
                else:
                    color = BLACK if self.isDark(row - 1, col - 1) else WHITE
                    cprint(color, end="")
            print()
        print()

    def make(self):
        self._make(False, self._getBestMaskPattern())

    def _getBestMaskPattern(self):
        minLostPoint = 0
        pattern = 0
        for i in range(8):
            self._make(True, i)
            lostPoint = QRUtil.getLostPoint(self)
            if i == 0 or minLostPoint > lostPoint:
                minLostPoint = lostPoint
                pattern = i
        return pattern

    def _make(self, test, maskPattern):
        self.moduleCount = self.typeNumber * 4 + 17
        self.modules = [[None] * self.moduleCount
                        for i in range(self.moduleCount)]

        self._setupPositionProbePattern(0, 0)
        self._setupPositionProbePattern(self.moduleCount - 7, 0)
        self._setupPositionProbePattern(0, self.moduleCount - 7)

        self._setupPositionAdjustPattern()
        self._setupTimingPattern()

        self._setupTypeInfo(test, maskPattern)

        if self.typeNumber >= 7:
            self._setupTypeNumber(test)

        data = QRCode._createData(
            self.typeNumber,
            self.errorCorrectLevel,
            self.qrDataList)

        self._mapData(data, maskPattern)

    def _mapData(self, data, maskPattern):
        rows = list(range(self.moduleCount))
        cols = [col - 1 if col <= 6 else col
                for col in range(self.moduleCount - 1, 0, -2)]
        maskFunc = QRUtil.getMaskFunction(maskPattern)

        byteIndex = 0
        bitIndex = 7

        for col in cols:
            rows.reverse()
            for row in rows:
                for c in range(2):
                    if self.modules[row][col - c] is None:

                        dark = False
                        if byteIndex < len(data):
                            dark = ((data[byteIndex] >> bitIndex) & 1) == 1
                        if maskFunc(row, col - c):
                            dark = not dark
                        self.modules[row][col - c] = dark

                        bitIndex -= 1
                        if bitIndex == -1:
                            byteIndex += 1
                            bitIndex = 7

    def _setupPositionAdjustPattern(self):
        pos = QRUtil.getPatternPosition(self.typeNumber)
        for row in pos:
            for col in pos:
                if self.modules[row][col] is not None:
                    continue
                for r in range(-2, 3):
                    for c in range(-2, 3):
                        self.modules[row + r][col + c] = (
                            r == -2 or r == 2 or c == -2 or c == 2
                            or (r == 0 and c == 0))

    def _setupPositionProbePattern(self, row, col):
        for r in range(-1, 8):
            for c in range(-1, 8):
                if (row + r <= -1 or self.moduleCount <= row + r
                        or col + c <= -1 or self.moduleCount <= col + c):
                    continue
                self.modules[row + r][col + c] = (
                    (0 <= r and r <= 6 and (c == 0 or c == 6))
                    or (0 <= c and c <= 6 and (r == 0 or r == 6))
                    or (2 <= r and r <= 4 and 2 <= c and c <= 4))

    def _setupTimingPattern(self):
        for r in range(8, self.moduleCount - 8):
            if self.modules[r][6] is not None:
                continue
            self.modules[r][6] = r % 2 == 0
        for c in range(8, self.moduleCount - 8):
            if self.modules[6][c] is not None:
                continue
            self.modules[6][c] = c % 2 == 0

    def _setupTypeNumber(self, test):
        bits = QRUtil.getBCHTypeNumber(self.typeNumber)
        for i in range(18):
            self.modules[i // 3][i % 3 + self.moduleCount - 8 - 3] = (
                not test and ((bits >> i) & 1) == 1)
        for i in range(18):
            self.modules[i % 3 + self.moduleCount - 8 - 3][i // 3] = (
                not test and ((bits >> i) & 1) == 1)

    def _setupTypeInfo(self, test, maskPattern):
        data = (self.errorCorrectLevel << 3) | maskPattern
        bits = QRUtil.getBCHTypeInfo(data)

        # vertical
        for i in range(15):
            mod = not test and ((bits >> i) & 1) == 1
            if i < 6:
                self.modules[i][8] = mod
            elif i < 8:
                self.modules[i + 1][8] = mod
            else:
                self.modules[self.moduleCount - 15 + i][8] = mod

        # horizontal
        for i in range(15):
            mod = not test and ((bits >> i) & 1) == 1
            if i < 8:
                self.modules[8][self.moduleCount - i - 1] = mod
            elif i < 9:
                self.modules[8][15 - i - 1 + 1] = mod
            else:
                self.modules[8][15 - i - 1] = mod

        # fixed
        self.modules[self.moduleCount - 8][8] = not test

    @staticmethod
    def _createData(typeNumber, errorCorrectLevel, dataArray):
        rsBlocks = RSBlock.getRSBlocks(typeNumber, errorCorrectLevel)

        buffer = BitBuffer()

        for data in dataArray:
            buffer.put(data.getMode(), 4)
            buffer.put(data.getLength(), data.getLengthInBits(typeNumber))
            data.write(buffer)

        totalDataCount = sum(rsBlock.getDataCount()
                             for rsBlock in rsBlocks)

        if buffer.getLengthInBits() > totalDataCount * 8:
            raise Exception('code length overflow. (%s>%s)' %
                            (buffer.getLengthInBits(), totalDataCount * 8))

        # end code
        if buffer.getLengthInBits() + 4 <= totalDataCount * 8:
            buffer.put(0, 4)

        # padding
        while buffer.getLengthInBits() % 8 != 0:
            buffer.put(False)

        # padding
        while True:
            if buffer.getLengthInBits() >= totalDataCount * 8:
                break
            buffer.put(QRCode.PAD0, 8)
            if buffer.getLengthInBits() >= totalDataCount * 8:
                break
            buffer.put(QRCode.PAD1, 8)

        return QRCode._createBytes(buffer, rsBlocks)

    @staticmethod
    def _createBytes(buffer, rsBlocks):
        offset = 0

        maxDcCount = 0
        maxEcCount = 0

        dcdata = [None] * len(rsBlocks)
        ecdata = [None] * len(rsBlocks)

        for r in range(len(rsBlocks)):

            dcCount = rsBlocks[r].getDataCount()
            ecCount = rsBlocks[r].getTotalCount() - dcCount

            maxDcCount = max(maxDcCount, dcCount)
            maxEcCount = max(maxEcCount, ecCount)

            dcdata[r] = [0] * dcCount
            for i in range(len(dcdata[r])):
                dcdata[r][i] = 0xff & buffer.getBuffer()[i + offset]
            offset += dcCount

            rsPoly = QRUtil.getErrorCorrectPolynomial(ecCount)
            rawPoly = Polynomial(dcdata[r], rsPoly.getLength() - 1)

            modPoly = rawPoly.mod(rsPoly)
            ecdata[r] = [0] * (rsPoly.getLength() - 1)
            for i in range(len(ecdata[r])):
                modIndex = i + modPoly.getLength() - len(ecdata[r])
                ecdata[r][i] = modPoly.get(modIndex) if modIndex >= 0 else 0

        totalCodeCount = sum(rsBlock.getTotalCount()
                             for rsBlock in rsBlocks)

        data = [0] * totalCodeCount

        index = 0

        for i in range(maxDcCount):
            for r in range(len(rsBlocks)):
                if i < len(dcdata[r]):
                    data[index] = dcdata[r][i]
                    index += 1

        for i in range(maxEcCount):
            for r in range(len(rsBlocks)):
                if i < len(ecdata[r]):
                    data[index] = ecdata[r][i]
                    index += 1

        return data

    @staticmethod
    def getMinimumQRCode(data, errorCorrectLevel):
        mode = Mode.MODE_8BIT_BYTE  # fixed to 8bit byte
        qr = QRCode()
        qr.setErrorCorrectLevel(errorCorrectLevel)
        qr.addData(data)
        length = qr.getData(0).getLength()
        for typeNumber in range(1, 11):
            if length <= QRUtil.getMaxLength(
                    typeNumber, mode, errorCorrectLevel):
                qr.setTypeNumber(typeNumber)
                break
        qr.make()
        return qr


class Mode:
    MODE_NUMBER = 1 << 0
    MODE_ALPHA_NUM = 1 << 1
    MODE_8BIT_BYTE = 1 << 2
    MODE_KANJI = 1 << 3


class ErrorCorrectLevel:
    L = 1  # 7%
    M = 0  # 15%
    Q = 3  # 25%
    H = 2  # 30%


class MaskPattern:
    PATTERN000 = 0
    PATTERN001 = 1
    PATTERN010 = 2
    PATTERN011 = 3
    PATTERN100 = 4
    PATTERN101 = 5
    PATTERN110 = 6
    PATTERN111 = 7


class QRUtil:

    @staticmethod
    def getPatternPosition(typeNumber):
        return QRUtil.PATTERN_POSITION_TABLE[typeNumber - 1]

    PATTERN_POSITION_TABLE = [
        [],
        [6, 18],
        [6, 22],
        [6, 26],
        [6, 30],
        [6, 34],
        [6, 22, 38],
        [6, 24, 42],
        [6, 26, 46],
        [6, 28, 50],
        [6, 30, 54],
        [6, 32, 58],
        [6, 34, 62],
        [6, 26, 46, 66],
        [6, 26, 48, 70],
        [6, 26, 50, 74],
        [6, 30, 54, 78],
        [6, 30, 56, 82],
        [6, 30, 58, 86],
        [6, 34, 62, 90],
        [6, 28, 50, 72, 94],
        [6, 26, 50, 74, 98],
        [6, 30, 54, 78, 102],
        [6, 28, 54, 80, 106],
        [6, 32, 58, 84, 110],
        [6, 30, 58, 86, 114],
        [6, 34, 62, 90, 118],
        [6, 26, 50, 74, 98, 122],
        [6, 30, 54, 78, 102, 126],
        [6, 26, 52, 78, 104, 130],
        [6, 30, 56, 82, 108, 134],
        [6, 34, 60, 86, 112, 138],
        [6, 30, 58, 86, 114, 142],
        [6, 34, 62, 90, 118, 146],
        [6, 30, 54, 78, 102, 126, 150],
        [6, 24, 50, 76, 102, 128, 154],
        [6, 28, 54, 80, 106, 132, 158],
        [6, 32, 58, 84, 110, 136, 162],
        [6, 26, 54, 82, 110, 138, 166],
        [6, 30, 58, 86, 114, 142, 170]
    ]

    MAX_LENGTH = [
        [[41,  25,  17,  10],  [34,  20,  14,  8],   [
            27,  16,  11,  7],  [17,  10,  7,   4]],
        [[77,  47,  32,  20],  [63,  38,  26,  16],
            [48,  29,  20,  12], [34,  20,  14,  8]],
        [[127, 77,  53,  32],  [101, 61,  42,  26],  [
            77,  47,  32,  20], [58,  35,  24,  15]],
        [[187, 114, 78,  48],  [149, 90,  62,  38],  [
            111, 67,  46,  28], [82,  50,  34,  21]],
        [[255, 154, 106, 65],  [202, 122, 84,  52],  [
            144, 87,  60,  37], [106, 64,  44,  27]],
        [[322, 195, 134, 82],  [255, 154, 106, 65],  [
            178, 108, 74,  45], [139, 84,  58,  36]],
        [[370, 224, 154, 95],  [293, 178, 122, 75],  [
            207, 125, 86,  53], [154, 93,  64,  39]],
        [[461, 279, 192, 118], [365, 221, 152, 93],  [
            259, 157, 108, 66], [202, 122, 84,  52]],
        [[552, 335, 230, 141], [432, 262, 180, 111], [
            312, 189, 130, 80], [235, 143, 98,  60]],
        [[652, 395, 271, 167], [513, 311, 213, 131],
            [364, 221, 151, 93], [288, 174, 119, 74]]
    ]

    @staticmethod
    def getMaxLength(typeNumber, mode, errorCorrectLevel):
        t = typeNumber - 1
        e = {
            ErrorCorrectLevel.L: 0,
            ErrorCorrectLevel.M: 1,
            ErrorCorrectLevel.Q: 2,
            ErrorCorrectLevel.H: 3
        }[errorCorrectLevel]
        m = {
            Mode.MODE_NUMBER: 0,
            Mode.MODE_ALPHA_NUM: 1,
            Mode.MODE_8BIT_BYTE: 2,
            Mode.MODE_KANJI: 3
        }[mode]
        return QRUtil.MAX_LENGTH[t][e][m]

    @staticmethod
    def getErrorCorrectPolynomial(errorCorrectLength):
        a = Polynomial([1])
        for i in range(errorCorrectLength):
            a = a.multiply(Polynomial([1, QRMath.gexp(i)]))
        return a

    @staticmethod
    def getMaskFunction(maskPattern):
        return {
            MaskPattern.PATTERN000:
                lambda i, j: (i + j) % 2 == 0,
            MaskPattern.PATTERN001:
                lambda i, j: i % 2 == 0,
            MaskPattern.PATTERN010:
                lambda i, j: j % 3 == 0,
            MaskPattern.PATTERN011:
                lambda i, j: (i + j) % 3 == 0,
            MaskPattern.PATTERN100:
                lambda i, j: (i // 2 + j // 3) % 2 == 0,
            MaskPattern.PATTERN101:
                lambda i, j: (i * j) % 2 + (i * j) % 3 == 0,
            MaskPattern.PATTERN110:
                lambda i, j: ((i * j) % 2 + (i * j) % 3) % 2 == 0,
            MaskPattern.PATTERN111:
                lambda i, j: ((i * j) % 3 + (i + j) % 2) % 2 == 0
        }[maskPattern]

    @staticmethod
    def getLostPoint(qrcode):
        moduleCount = qrcode.getModuleCount()
        lostPoint = 0

        # LEVEL1
        for row in range(moduleCount):
            for col in range(moduleCount):
                sameCount = 0
                dark = qrcode.isDark(row, col)
                for r in range(-1, 2):
                    if row + r < 0 or moduleCount <= row + r:
                        continue
                    for c in range(-1, 2):
                        if col + c < 0 or moduleCount <= col + c:
                            continue
                        if r == 0 and c == 0:
                            continue
                        if dark == qrcode.isDark(row + r, col + c):
                            sameCount += 1
                if sameCount > 5:
                    lostPoint += (3 + sameCount - 5)

        # LEVEL2
        for row in range(moduleCount - 1):
            for col in range(moduleCount - 1):
                count = 0
                if qrcode.isDark(row, col):
                    count += 1
                if qrcode.isDark(row + 1, col):
                    count += 1
                if qrcode.isDark(row, col + 1):
                    count += 1
                if qrcode.isDark(row + 1, col + 1):
                    count += 1
                if count == 0 or count == 4:
                    lostPoint += 3

        # LEVEL3
        for row in range(moduleCount):
            for col in range(moduleCount - 6):
                if (qrcode.isDark(row, col)
                        and not qrcode.isDark(row, col + 1)
                        and qrcode.isDark(row, col + 2)
                        and qrcode.isDark(row, col + 3)
                        and qrcode.isDark(row, col + 4)
                        and not qrcode.isDark(row, col + 5)
                        and qrcode.isDark(row, col + 6)):
                    lostPoint += 40

        for col in range(moduleCount):
            for row in range(moduleCount - 6):
                if (qrcode.isDark(row, col)
                        and not qrcode.isDark(row + 1, col)
                        and qrcode.isDark(row + 2, col)
                        and qrcode.isDark(row + 3, col)
                        and qrcode.isDark(row + 4, col)
                        and not qrcode.isDark(row + 5, col)
                        and qrcode.isDark(row + 6, col)):
                    lostPoint += 40

        # LEVEL4
        darkCount = 0
        for col in range(moduleCount):
            for row in range(moduleCount):
                if qrcode.isDark(row, col):
                    darkCount += 1

        ratio = abs(100 * darkCount // moduleCount // moduleCount - 50) // 5
        lostPoint += ratio * 10

        return lostPoint

    G15 = ((1 << 10) | (1 << 8) | (1 << 5) | (1 << 4) |
           (1 << 2) | (1 << 1) | (1 << 0))
    G18 = ((1 << 12) | (1 << 11) | (1 << 10) | (1 << 9) |
           (1 << 8) | (1 << 5) | (1 << 2) | (1 << 0))
    G15_MASK = (1 << 14) | (1 << 12) | (1 << 10) | (1 << 4) | (1 << 1)

    @staticmethod
    def getBCHTypeInfo(data):
        d = data << 10
        while QRUtil.getBCHDigit(d) - QRUtil.getBCHDigit(QRUtil.G15) >= 0:
            d ^= (QRUtil.G15 << (QRUtil.getBCHDigit(d) -
                                 QRUtil.getBCHDigit(QRUtil.G15)))
        return ((data << 10) | d) ^ QRUtil.G15_MASK

    @staticmethod
    def getBCHTypeNumber(data):
        d = data << 12
        while QRUtil.getBCHDigit(d) - QRUtil.getBCHDigit(QRUtil.G18) >= 0:
            d ^= (QRUtil.G18 << (QRUtil.getBCHDigit(d) -
                                 QRUtil.getBCHDigit(QRUtil.G18)))
        return (data << 12) | d

    @staticmethod
    def getBCHDigit(data):
        digit = 0
        while data != 0:
            digit += 1
            data >>= 1
        return digit

    @staticmethod
    def stringToBytes(s):
        return [ord(c) & 0xff for c in s]


class QR8BitByte:

    def __init__(self, data):
        self.mode = Mode.MODE_8BIT_BYTE
        self.data = data

    def getMode(self):
        return self.mode

    def getData(self):
        return self.data

    '''
    def write(self, buffer): raise Exception('not implemented.')
    def getLength(self): raise Exception('not implemented.')
    '''

    def write(self, buffer):
        data = QRUtil.stringToBytes(self.getData())
        for d in data:
            buffer.put(d, 8)

    def getLength(self):
        return len(QRUtil.stringToBytes(self.getData()))

    def getLengthInBits(self, type):
        if 1 <= type and type < 10:  # 1 - 9
            return {
                Mode.MODE_NUMBER:    10,
                Mode.MODE_ALPHA_NUM: 9,
                Mode.MODE_8BIT_BYTE: 8,
                Mode.MODE_KANJI:     8
            }[self.mode]

        elif type < 27:  # 10 - 26
            return {
                Mode.MODE_NUMBER:    12,
                Mode.MODE_ALPHA_NUM: 11,
                Mode.MODE_8BIT_BYTE: 16,
                Mode.MODE_KANJI:     10
            }[self.mode]

        elif type < 41:  # 27 - 40
            return {
                Mode.MODE_NUMBER:    14,
                Mode.MODE_ALPHA_NUM: 13,
                Mode.MODE_8BIT_BYTE: 16,
                Mode.MODE_KANJI:     12
            }[self.mode]

        else:
            raise Exception('type:%s' % type)


class QRMath:

    EXP_TABLE = None
    LOG_TABLE = None

    @staticmethod
    def _init():
        QRMath.EXP_TABLE = [0] * 256
        for i in range(256):
            QRMath.EXP_TABLE[i] = (1 << i if i < 8 else
                                   QRMath.EXP_TABLE[i - 4]
                                   ^ QRMath.EXP_TABLE[i - 5]
                                   ^ QRMath.EXP_TABLE[i - 6]
                                   ^ QRMath.EXP_TABLE[i - 8])

        QRMath.LOG_TABLE = [0] * 256
        for i in range(255):
            QRMath.LOG_TABLE[QRMath.EXP_TABLE[i]] = i

    @staticmethod
    def glog(n):
        if n < 1:
            raise Exception('log(%s)' % n)
        return QRMath.LOG_TABLE[n]

    @staticmethod
    def gexp(n):
        while n < 0:
            n += 255
        while n >= 256:
            n -= 255
        return QRMath.EXP_TABLE[n]


# initialize statics
QRMath._init()


class Polynomial:

    def __init__(self, num, shift=0):
        offset = 0
        length = len(num)
        while offset < length and num[offset] == 0:
            offset += 1
        self.num = num[offset:] + [0] * shift

    def get(self, index):
        return self.num[index]

    def getLength(self):
        return len(self.num)

    def __repr__(self):
        return ','.join([str(self.get(i))
                         for i in range(self.getLength())])

    def toLogString(self):
        return ','.join([str(QRMath.glog(self.get(i)))
                         for i in range(self.getLength())])

    def multiply(self, e):
        num = [0] * (self.getLength() + e.getLength() - 1)
        for i in range(self.getLength()):
            for j in range(e.getLength()):
                num[i + j] ^= QRMath.gexp(QRMath.glog(self.get(i)) +
                                          QRMath.glog(e.get(j)))
        return Polynomial(num)

    def mod(self, e):
        if self.getLength() - e.getLength() < 0:
            return self
        ratio = QRMath.glog(self.get(0)) - QRMath.glog(e.get(0))
        num = self.num[:]
        for i in range(e.getLength()):
            num[i] ^= QRMath.gexp(QRMath.glog(e.get(i)) + ratio)
        return Polynomial(num).mod(e)


class RSBlock:

    RS_BLOCK_TABLE = [

        # L
        # M
        # Q
        # H

        # 1
        [1, 26, 19],
        [1, 26, 16],
        [1, 26, 13],
        [1, 26, 9],

        # 2
        [1, 44, 34],
        [1, 44, 28],
        [1, 44, 22],
        [1, 44, 16],

        # 3
        [1, 70, 55],
        [1, 70, 44],
        [2, 35, 17],
        [2, 35, 13],

        # 4
        [1, 100, 80],
        [2, 50, 32],
        [2, 50, 24],
        [4, 25, 9],

        # 5
        [1, 134, 108],
        [2, 67, 43],
        [2, 33, 15, 2, 34, 16],
        [2, 33, 11, 2, 34, 12],

        # 6
        [2, 86, 68],
        [4, 43, 27],
        [4, 43, 19],
        [4, 43, 15],

        # 7
        [2, 98, 78],
        [4, 49, 31],
        [2, 32, 14, 4, 33, 15],
        [4, 39, 13, 1, 40, 14],

        # 8
        [2, 121, 97],
        [2, 60, 38, 2, 61, 39],
        [4, 40, 18, 2, 41, 19],
        [4, 40, 14, 2, 41, 15],

        # 9
        [2, 146, 116],
        [3, 58, 36, 2, 59, 37],
        [4, 36, 16, 4, 37, 17],
        [4, 36, 12, 4, 37, 13],

        # 10
        [2, 86, 68, 2, 87, 69],
        [4, 69, 43, 1, 70, 44],
        [6, 43, 19, 2, 44, 20],
        [6, 43, 15, 2, 44, 16]
    ]

    def __init__(self, totalCount, dataCount):
        self.totalCount = totalCount
        self.dataCount = dataCount

    def getDataCount(self):
        return self.dataCount

    def getTotalCount(self):
        return self.totalCount

    def __repr__(self):
        return ('(total=%s,data=%s)' % (self.totalCount, self.dataCount))

    @staticmethod
    def getRSBlocks(typeNumber, errorCorrectLevel):
        rsBlock = RSBlock.getRsBlockTable(typeNumber, errorCorrectLevel)
        length = len(rsBlock) // 3
        list = []
        for i in range(length):
            count = rsBlock[i * 3 + 0]
            totalCount = rsBlock[i * 3 + 1]
            dataCount = rsBlock[i * 3 + 2]
            list += [RSBlock(totalCount, dataCount)] * count
        return list

    @staticmethod
    def getRsBlockTable(typeNumber, errorCorrectLevel):
        return {
            ErrorCorrectLevel.L:
                RSBlock.RS_BLOCK_TABLE[(typeNumber - 1) * 4 + 0],
            ErrorCorrectLevel.M:
                RSBlock.RS_BLOCK_TABLE[(typeNumber - 1) * 4 + 1],
            ErrorCorrectLevel.Q:
                RSBlock.RS_BLOCK_TABLE[(typeNumber - 1) * 4 + 2],
            ErrorCorrectLevel.H:
                RSBlock.RS_BLOCK_TABLE[(typeNumber - 1) * 4 + 3]
        }[errorCorrectLevel]


class BitBuffer:

    def __init__(self, inclements=32):
        self.inclements = inclements
        self.buffer = [0] * self.inclements
        self.length = 0

    def getBuffer(self):
        return self.buffer

    def getLengthInBits(self):
        return self.length

    def get(self, index):
        return ((self.buffer[index // 8] >> (7 - index % 8)) & 1) == 1

    def putBit(self, bit):
        if self.length == len(self.buffer) * 8:
            self.buffer += [0] * self.inclements
        if bit:
            self.buffer[self.length // 8] |= (0x80 >> (self.length % 8))
        self.length += 1

    def put(self, num, length):
        for i in range(length):
            self.putBit(((num >> (length - i - 1)) & 1) == 1)

    def __repr__(self):
        return ''.join('1' if self.get(i) else '0'
                       for i in range(self.getLengthInBits()))



"""Utility functions for Airshare."""


import mimetypes
import os
import pyperclip
import re
import socket
import tempfile
from time import strftime
from zipfile import ZipFile
from zeroconf import IPVersion, ServiceInfo, Zeroconf


from .qrcode import ErrorCorrectLevel, QRCode


__all__ = ["get_local_ip_address", "qr_code", "get_service_info",
           "register_service", "get_zip_file", "unzip_file",
           "get_clipboard_paths", "is_file_copyable"]


# Local IP Address


def get_local_ip_address():
    r"""Obtains the device's local network IP address.

    Returns
    -------
    ip : bytes
        Packed 32-bit representation of the device's local IP Address.
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("10.255.255.255", 1))
    ip = s.getsockname()[0]
    s.close()
    ip = socket.inet_aton(ip)
    return ip


# QR Code Utility


def qr_code(url):
    r"""Generate QR Code from URL and print it.

    Parameters
    ----------
    url : str
        URL to create the QR Code for.
    """
    qr = QRCode.getMinimumQRCode(url, ErrorCorrectLevel.M)
    qr.setErrorCorrectLevel(ErrorCorrectLevel.M)
    qr.make()
    qr.printQr()


# Zeroconf Utilities


def get_service_info(code):
    r"""Get service information for an Airshare service.

    Parameters
    ----------
    code : str
        Identifying code for the Airshare service.

    Returns
    -------
    info : zeroconf.ServiceInfo
        Details of the Airshare service.
    """
    zeroconf = Zeroconf(ip_version=IPVersion.V4Only)
    service = "_airshare._http._tcp.local."
    info = zeroconf.get_service_info(service, code + service)
    return info


def register_service(code, addresses, port):
    r"""Registers an Airshare Multicast-DNS service based in the local network.

    Parameters
    ----------
    code : str
        Identifying code for the Airshare service.
    addresses : list
        List of local network IP Addresses for the service.
    port : int
        Port number for the Airshare service's server.

    Returns
    -------
    info : zeroconf.ServiceInfo
        Details of the Airshare service.
    """
    zeroconf = Zeroconf(ip_version=IPVersion.V4Only)
    service = "_airshare._http._tcp.local."
    info = ServiceInfo(
        service,
        code + service,
        addresses=addresses,
        port=port,
        server=code + ".local."
    )
    zeroconf.register_service(info)
    return info


# Zip and Unzip


def get_zip_file(files):
    r"""Creates a temporary Zip Archive of files and directories.

    Parameters
    ----------
    files : list
        List of paths of files and directories to compress.

    Returns
    -------
    zip_file_path : str
        Canonical file path of the temporary Zip Archive file.
    zip_file_name : str
        File name to be assigned to the Zip Archive (during sending).
    """
    files = [os.path.realpath(x) for x in files]
    _, zip_file_path = tempfile.mkstemp(prefix="airshare", suffix=".zip")
    zip_archive = ZipFile(zip_file_path, "w")
    num_files = len(files)
    index = -1
    if num_files == 1:
        index = 0
    for item in files:
        index += len(item.split(os.path.sep))
        if os.path.isdir(item):
            for root, _, file_list in os.walk(item):
                for file in file_list:
                    file_path = os.path.realpath(os.path.join(root, file))
                    zip_archive.write(file_path, os.path.join(
                        *tuple(root.split(os.path.sep)[index:] + [file])))
        else:
            file_path = os.path.realpath(item)
            zip_archive.write(file_path, os.path.join(
                *tuple(file_path.split(os.path.sep)[index:])))
        index = -1
    zip_archive.close()
    zip_file_path = os.path.abspath(zip_file_path)
    zip_file_name = "airshare.zip"
    if num_files == 1:
        zip_file_name = os.path.splitext(
            os.path.realpath(files[0]).split(os.path.sep)[-1])[0] + ".zip"
    return zip_file_path, zip_file_name


def unzip_file(zip_file_path):
    r"""Unzips a Zip Archive file into a new directory.

    Parameters
    ----------
    zip_file_path : str
        Path of the Zip Archive file to unzip.

    Returns
    -------
    zip_dir : str
        Canonical path of the unzipped directory.
    """
    zip_dir = zip_file_path[:-4]
    if os.path.isdir(zip_dir):
        zip_dir += "-" + strftime("%Y%m%d%H%M%S")
    os.mkdir(zip_dir)
    with ZipFile(zip_file_path, "r") as zip_archive:
        zip_archive.extractall(zip_dir)
    zip_dir = os.path.realpath(zip_dir)
    return zip_dir


# Clipboard Utilities


def get_clipboard_paths():
    r"""Extract file paths from the clipboard.

    Returns
    -------
    file_paths : list
        List of canonical paths extracted from the clipboard.
    """
    file_paths = []
    clipboard = pyperclip.paste()
    erase = ["x-special/nautilus-clipboard\ncopy\n", "file://", "\r", "'", '"']
    file_paths = re.sub("|".join(erase), "", clipboard.strip()).split("\n")
    file_paths = [os.path.realpath(str(x).strip()) for x in file_paths]
    return file_paths


def is_file_copyable(file_path):
    r"""Check if a file can be copied to the clipboard or not.

    Parameters
    ----------
    file_path : str
        Path of the file to check.

    Returns
    -------
    copyable : boolean
        True if the file can be copied to the clipboard, False otherwise.
    """
    file_type = mimetypes.guess_type(file_path)[0]
    copyable = False
    if file_type is not None:
        if (re.findall("text|json", file_type, re.IGNORECASE)):
            copyable = True
        else:
            copyable = False
    return copyable



"""Module for receiving data and hosting receiving servers."""


from aiohttp import web
import asyncio
import humanize
from multiprocessing import Process
import os
import pkgutil
import platform
import requests
import socket
import sys
from time import sleep, strftime
from tqdm import tqdm
from zipfile import is_zipfile


from .exception import CodeExistsError, CodeNotFoundError, IsNotSenderError
from .utils import get_local_ip_address, get_service_info, qr_code, \
    register_service, unzip_file


__all__ = ["receive", "receive_server", "receive_server_proc"]


# Request handlers


async def _upload_page(request):
    """Renders an upload page. GET handler for route '/'."""
    upload = pkgutil.get_data(__name__, "static/upload.html").decode()
    return web.Response(text=upload, content_type="text/html")


async def _uploaded_file_receiver(request):
    """Receives an uploaded file. POST handler for '/upload'."""
    progress_queue = request.app["progress_queue"]
    tqdm_position = await progress_queue.get()
    decompress = request.app["decompress"]
    compress_header = request.headers.get("airshare-compress") or "false"
    if compress_header == "true":
        decompress = True
    total = 0
    reader = await request.multipart()
    field = await reader.next()
    file_name = field.filename.replace("'", "")
    file_path = os.getcwd() + os.path.sep + file_name
    if os.path.isfile(file_path):
        file_name, file_ext = os.path.splitext(file_name)
        file_name = file_name + "-" + strftime("%Y%m%d%H%M%S") + file_ext
        file_path = os.getcwd() + os.path.sep + file_name
    desc = "Downloading `" + file_name + "`"
    bar = tqdm(desc=desc, total=None, unit="B", unit_scale=1,
               position=tqdm_position, leave=False)
    with open(file_path, "wb") as f:
        while True:
            chunk = await field.read_chunk()
            if not chunk:
                break
            total += len(chunk)
            f.write(chunk)
            bar.update(len(chunk))
    await progress_queue.put(tqdm_position)
    if is_zipfile(file_path) and decompress:
        zip_dir = unzip_file(file_path)
        tqdm.write("Downloaded and decompressed to `" + zip_dir + "`!")
        os.remove(file_path)
    else:
        tqdm.write("Downloaded `" + file_name + "`!")
    file_name = field.filename
    file_size = humanize.naturalsize(total)
    text = "{} ({}) successfully received!".format(file_name, file_size)
    return web.Response(text=text)


async def _is_airshare_upload_receiver(request):
    """Returns 'Upload Receiver'. GET handler for '/airshare'."""
    return web.Response(text="Upload Receiver")


# Receiver functions


def receive(*, code, decompress=False):
    r"""Receive file(s) from a sending server.

    Parameters
    ----------
    code : str
        Identifying code for the Airshare sending server.
    decompress : boolean, default=False
        Flag to enable or disable decompression (Zip).

    Returns
    -------
    text (or) file_path : str
        Returns the text or path of the file received, if successful.
    """
    info = get_service_info(code)
    if info is None:
        raise CodeNotFoundError(code)
    ip = socket.inet_ntoa(info.addresses[0])
    url = "http://" + ip + ":" + str(info.port)
    airshare_type = requests.get(url + "/airshare").text
    if "Sender" not in airshare_type:
        raise IsNotSenderError(code)
    print("Receiving from Airshare `" + code + ".local`...")
    sleep(2)
    if airshare_type == "Text Sender":
        text = requests.get(url + "/text").text
        print("Received: " + text)
        return text
    elif airshare_type == "File Sender":
        with requests.get(url + "/download", stream=True) as r:
            r.raise_for_status()
            header = r.headers["content-disposition"]
            compress_header = r.headers.get("airshare-compress") or "false"
            if compress_header == "true":
                decompress = True
            file_name = header.split("; ")[1].split("=")[1] \
                              .replace("'", "")
            file_path = os.getcwd() + os.path.sep + file_name
            file_size = int(header.split("=")[-1])
            if os.path.isfile(file_path):
                file_name, file_ext = os.path.splitext(file_name)
                file_name += "-" + strftime("%Y%m%d%H%M%S") + file_ext
                file_path = os.getcwd() + os.path.sep + file_name
            with open(file_path, "wb") as f:
                desc = "Downloading `" + file_name + "`"
                bar = tqdm(desc=desc, total=file_size, unit="B",
                           unit_scale=1, leave=False)
                for chunk in r.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)
                        bar.update(len(chunk))
            file_path = os.path.realpath(file_path)
            if is_zipfile(file_path) and decompress:
                zip_dir = unzip_file(file_path)
                tqdm.write("Downloaded and decompressed to `" + zip_dir + "`!")
                os.remove(file_path)
                file_path = zip_dir
            else:
                tqdm.write("Downloaded `" + file_path + "`!")
            return file_path


def receive_server(*, code, decompress=False, port=80):
    r"""Serves a file receiver and registers it as a Multicast-DNS service.

    Parameters
    ----------
    code : str
        Identifying code for the Airshare service and server.
    decompress : boolean, default=False
        Flag to enable or disable decompression (Zip).
    port : int, default=80
        Port number at which the server is hosted on the device.
    """
    info = get_service_info(code)
    if info is not None:
        raise CodeExistsError(code)
    addresses = [get_local_ip_address()]
    register_service(code, addresses, port)
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    app = web.Application()
    app["decompress"] = decompress
    app["progress_queue"] = asyncio.Queue()
    for pos in range(5):
        app["progress_queue"].put_nowait(pos)
    app.router.add_get(path="/", handler=_upload_page)
    app.router.add_get(path="/airshare", handler=_is_airshare_upload_receiver)
    app.router.add_post(path="/upload", handler=_uploaded_file_receiver)
    runner = web.AppRunner(app)
    loop.run_until_complete(runner.setup())
    site = web.TCPSite(runner, "0.0.0.0", str(port))
    loop.run_until_complete(site.start())
    url_port = ""
    if port != 80:
        url_port = ":" + str(port)
    ip = socket.inet_ntoa(addresses[0]) + url_port
    quit_msg = "`, press Ctrl+C to stop receiving..."
    if platform.system() == "Windows" and sys.version_info < (3, 8):
        quit_msg = "`, press Ctrl+Break to stop receiving..."
    print("Waiting for uploaded files at " + ip + " and `http://"
          + code + ".local" + url_port + quit_msg)
    qr_code("http://" + ip)
    if decompress:
        print("Note: Any Zip Archives will be decompressed!")
    loop.run_forever()


def receive_server_proc(*, code, decompress=False, port=80):
    r"""Creates a process with 'receive_server' as the target.

    Parameters
    ----------
    code : str
        Identifying code for the Airshare service and server.
    decompress : boolean, default=False
        Flag to enable or disable decompression (Zip).
    port : int, default=80
        Port number at which the server is hosted on the device.

    Returns
    -------
    process: multiprocessing.Process
        A multiprocessing.Process object with 'receive_server' as target.
    """
    kwargs = {"code": code, "decompress": decompress, "port": port}
    process = Process(target=receive_server, kwargs=kwargs)
    return process
